﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;
using Commons;
using System.Threading;
using System.Xml;
using System.Diagnostics;

namespace Zerich.Micex
{
    /// <summary>
    /// Адаптер API шлюза ММВБ
    /// </summary>
    class MicexApi
    {
        MicexSubType sub_type;
        List<string> auth_strings = new List<string>();

        Mutex tables_mutex = new Mutex();
        bool is_refreshing_quotes = false;
        bool is_refreshing_other_tables = false;

        string last_error;
        public string LastError { get { return last_error; } }

        Queue<List<Quote>> quotes = new Queue<List<Quote>>();
        public Queue<List<Quote>> QuotesQueue { get { return quotes; } }

        Queue<MicexOrderUpdate> orders = new Queue<MicexOrderUpdate>();
        public Queue<MicexOrderUpdate> OrderUpdates { get { return orders; } }

        Queue<MicexTradeUpdate> trades = new Queue<MicexTradeUpdate>();
        public Queue<MicexTradeUpdate> TradeUpdates { get { return trades; } }

        Queue<MicesSecuritiesUpdate> securities = new Queue<MicesSecuritiesUpdate>();
        public Queue<MicesSecuritiesUpdate> SecuritiesUpdates { get { return securities; } }

        List<int> instrument_tables = new List<int>();
        List<int> securities_tables = new List<int>();
        List<int> orders_tables = new List<int>();
        int trades_table;
        FieldsConverter fields_converter = new FieldsConverter();

        MicexConnectionsDeamon deamon = new MicexConnectionsDeamon();

        public MicexApi(XmlElement connections_xml_element, string micex_tables_structure_filename, MicexSubType sub_type)
        {
            foreach (XmlNode node in connections_xml_element.ChildNodes)
            {
                if (!(node is XmlElement) || node.Name != "Connection")
                    continue;
                XmlElement element = node as XmlElement;

                /// Формируем строку параметров
                StringBuilder sb = new StringBuilder();
                sb.AppendFormat("{0}={1}\r\n", "Host", Utils.FindValue<string>("Host", element));
                sb.AppendFormat("{0}={1}\r\n", "Server", Utils.FindValue<string>("Server", element));
                sb.AppendFormat("{0}={1}\r\n", "UserId", Utils.FindValue<string>("UserId", element));
                sb.AppendFormat("{0}={1}\r\n", "Password", Utils.FindValue<string>("Password", element));
                sb.AppendFormat("{0}={1}\r\n", "Interface", Utils.FindValue<string>("Interface", element));
                sb.Append("Logging=0");
                auth_strings.Add(sb.ToString());
            }

            this.sub_type = sub_type;
            fields_converter.Reload(micex_tables_structure_filename);
        }

        public void Connect()
        {
            deamon.OpenConnections(auth_strings);
            if (sub_type == MicexSubType.Spot)
                SetupSecurities("FOND", "EQBR");
            else
                SetupSecurities();
            //SetupSecurities("FOND", "EQNE");
            SetupOrders();
            SetupTrades();
        }

        public void Disconnect()
        {
            lock (deamon)
            {
                int connection_id = -1;
                try
                {
                    connection_id = deamon.GetConnectionId(MicexConnectionType.RefreshTables);
                    if (connection_id < 0)
                    {
                        Log.Push(LogItemType.Error, "Не удается получить ни одного соединения для закрытия подключения к шлюзу");
                        return;
                    }

                    foreach (int c in instrument_tables)
                        MTECloseTable(connection_id, c);

                    foreach (int c in orders_tables)
                        MTECloseTable(connection_id, c);

                    foreach (int c in securities_tables)
                        MTECloseTable(connection_id, c);

                    MTECloseTable(connection_id, trades_table);
                }
                catch (Exception e)
                {
                    Log.Push(e, "Ошибка при закрытии таблиц");
                }
                finally
                {
                    deamon.ReleaseConnectionId(connection_id);
                }

                deamon.CloseConnections();
            }
        }        

        public MicexReturnCode ExecTransaction(string name, FieldsetName fieldset_name, Dictionary<string, object> parameters, out string message, out TimeSpan gate_delay)
        {
            Stopwatch sw = Stopwatch.StartNew();
            Stopwatch total = Stopwatch.StartNew();

            int connection_id = -1;
            try
            {
//Log.Push(LogItemType.Debug, "Подготовка: {0} мс", sw.Elapsed.TotalMilliseconds);
//sw.Restart();
                connection_id = deamon.GetConnectionId(MicexConnectionType.PostOrders);
//Log.Push(LogItemType.Debug, "Получение соединения: {0} мс", sw.Elapsed.TotalMilliseconds);
//sw.Restart();
                string parameters_string = fields_converter.Encode(parameters, fieldset_name);
//Log.Push(LogItemType.Debug, "Кодирование параметров: {0} мс", sw.Elapsed.TotalMilliseconds);
//sw.Restart();
                byte[] bytes = new byte[255];
                Stopwatch stopwatch = Stopwatch.StartNew();
                int code = MTEExecTrans(connection_id, name, parameters_string, bytes);
                gate_delay = stopwatch.Elapsed;
//Log.Push(LogItemType.Debug, "Транзакция: {0} мс", sw.Elapsed.TotalMilliseconds);
//sw.Restart();
                message = ConvertAndDecode(bytes);
//Log.Push(LogItemType.Debug, "Формирование сообщения: {0} мс", sw.Elapsed.TotalMilliseconds);
//Log.Push(LogItemType.Debug, "Всего: {0} мс", total.Elapsed.TotalMilliseconds);

                return (MicexReturnCode)code;
            }
            catch (Exception e)
            {
                Log.Push(LogItemType.Error, "Ошибка при отправке транзакции в шлюз");
                Log.Push(LogItemType.Error, e.Message);
                message = "Ошибка при отправке транзакции в шлюз: " + e.Message;
                gate_delay = new TimeSpan();
                return MicexReturnCode.FatalError;
            }
            finally
            {
                deamon.ReleaseConnectionId(connection_id);
            }
        }

        private void SetupSecurities(string market = null, string board = null)
        {
            int connection_id = -1;
            try
            {
                connection_id = deamon.GetConnectionId(MicexConnectionType.RefreshTables);
                if (connection_id < 0)
                    throw new MicexApiException("Не получается получить идентификатор соединения");
                // Обновления securities
                // Формируем набор параметров для последующего кодирования
                Dictionary<string, object> ask_fields = new Dictionary<string, object>();
                if (sub_type == MicexSubType.Spot)
                {
                    ask_fields.Add("MARKETID", market);
                    ask_fields.Add("BOARDID", board);
                }

                // Превращаем набор параметров в строку, понятную шлюзу
                string parameters = fields_converter.Encode(ask_fields, FieldsetName.SecuritiesRequest);

                IntPtr[] raw_data = new IntPtr[1];
                byte[] message = new byte[255];
                int code = MTEOpenTable(connection_id, "SECURITIES", parameters, false, raw_data);

                ByteArrayReader reader = ProcessMTEResponse(raw_data);

                if (code < 0)
                {
                    last_error = ConvertAndDecode(reader.ReadBytesToEnd());
                    Log.Push(LogItemType.Error, "Не удалось подписаться на обновление securities");
                    return;
                }

                int Ref = reader.ReadInt();
                ProcessSecurities(reader);
                tables_mutex.WaitOne();
                securities_tables.Add(code);
                tables_mutex.ReleaseMutex();
                if (!String.IsNullOrEmpty(board))
                    Log.Push("Получение котировок по рынку {0}/{1} подключено", market, board);
                else
                    Log.Push("Получение котировок подключено");
            }
            catch (Exception e)
            {
                Log.Push(LogItemType.Error, "Ошибка при настройке подписки на securities");
                Log.Push(LogItemType.Error, e.Message);
            }
            finally
            {
                deamon.ReleaseConnectionId(connection_id);
            }
        }

        private void SetupOrders()
        {
            int connection_id = -1;
            try
            {
                connection_id = deamon.GetConnectionId(MicexConnectionType.RefreshTables);
                if (connection_id < 0)
                    throw new MicexApiException("Не получается получить идентификатор соединения");
                // Обновления ордеров
                IntPtr[] raw_data = new IntPtr[1];
                byte[] message = new byte[255];
                int code = MTEOpenTable(connection_id, "ORDERS", null, true, raw_data);

                ByteArrayReader reader = ProcessMTEResponse(raw_data);
                if (code < 0)
                {
                    last_error = ConvertAndDecode(reader.ReadBytesToEnd());
                    Log.Push(LogItemType.Error, "Не удалось подписаться на обновление списка ордеров");
                    return;
                }

                int Ref = reader.ReadInt();
                ProcessOrdersPack(reader);
                tables_mutex.WaitOne();
                orders_tables.Add(code);
                tables_mutex.ReleaseMutex();
                Log.Push("Получение обновлений по заявкам подключено");
            }
            catch (Exception e)
            {
                Log.Push(LogItemType.Error, "Ошибка при настройке подписки на ордера");
                Log.Push(LogItemType.Error, e.Message);
            }
            finally
            {
                deamon.ReleaseConnectionId(connection_id);
            }
        }

        private void SetupTrades()
        {
            int connection_id = -1;
            try
            {
                connection_id = deamon.GetConnectionId(MicexConnectionType.RefreshTables);
                if (connection_id < 0)
                    throw new MicexApiException("Не получается получить идентификатор соединения");
                // Обновления ордеров
                IntPtr[] raw_data = new IntPtr[1];
                byte[] message = new byte[255];
                int code = MTEOpenTable(connection_id, "TRADES", null, true, raw_data);

                ByteArrayReader reader = ProcessMTEResponse(raw_data);
                if (code < 0)
                {
                    last_error = ConvertAndDecode(reader.ReadBytesToEnd());
                    Log.Push(LogItemType.Error, "Не удалось подписаться на обновление списка сделок");
                    return;
                }

                int Ref = reader.ReadInt();
                ProcessTradesPack(reader);
                tables_mutex.WaitOne();
                trades_table = code;
                tables_mutex.ReleaseMutex();
                Log.Push("Получение обновлений по сделкам подключено");
            }
            catch (Exception e)
            {
                Log.Push(LogItemType.Error, "Ошибка при настройке подписки на сделки");
                Log.Push(LogItemType.Error, e.Message);
            }
            finally
            {
                deamon.ReleaseConnectionId(connection_id);
            }
        }
        
        /// <summary>
        /// Задает список инструментов, по которым выдаются котировки; уже существующий список обнуляется
        /// </summary>
        /// <param name="instruments"></param>
        public void SubscribeQuotes(IEnumerable<Instrument> instruments)
        {
            int connection_id = -1;
            try
            {
                connection_id = deamon.GetConnectionId(MicexConnectionType.RefreshTables);
                if (connection_id < 0)
                    throw new MicexApiException("Не получается получить идентификатор соединения");

                tables_mutex.WaitOne();
                foreach (int code in instrument_tables)
                    MTECloseTable(connection_id, code);

                List<Quote> pack = null;
                instrument_tables.Clear();
                tables_mutex.ReleaseMutex();

                foreach (Instrument i in instruments)
                {
                    // Формируем набор параметров для последующего кодирования
                    Dictionary<string, object> ask_fields = new Dictionary<string, object>();
                    ask_fields.Add("SECBOARD", i.ClassCode);
                    ask_fields.Add("SECCODE", i.Ticker);

                    // Превращаем набор параметров в строку, понятную шлюзу
                    string parameters = fields_converter.Encode(ask_fields, FieldsetName.OrderBookRequest);

                    IntPtr[] raw_data = new IntPtr[1];
                    byte[] message = new byte[255];
                    int code = MTEOpenTable(connection_id, "ORDERBOOK", parameters, false, raw_data);

                    ByteArrayReader reader = ProcessMTEResponse(raw_data);
                    if (code < 0)
                    {
                        last_error = ConvertAndDecode(reader.ReadBytesToEnd());
                        continue;
                    }

                    int Ref = reader.ReadInt();
                    ProcessQuotesPack(reader, ref pack);
                    tables_mutex.WaitOne();
                    instrument_tables.Add(code);
                    tables_mutex.ReleaseMutex();
                }

                quotes.Enqueue(pack);
            }
            catch (Exception e)
            {
                Log.Push(LogItemType.Error, "Ошибка при настройке подписки на котировки");
                Log.Push(LogItemType.Error, e.Message);
            }
            finally
            {
                deamon.ReleaseConnectionId(connection_id);
            }
        }

        /// <summary>
        /// Выполняет обновление таблиц и запись полученной информации в кеш
        /// </summary>
        public void RefreshQuotes()
        {
            // Обновляем котировки
            if (is_refreshing_quotes)
                return;

            is_refreshing_quotes = true;
            int connection_id = -1;
            try
            {
                connection_id = deamon.GetConnectionId(MicexConnectionType.RefreshTables);
                if (connection_id < 0)
                    throw new MicexApiException("Не получается получить идентификатор соединения");

                lock (instrument_tables)
                {
                    foreach (int code in instrument_tables)
                    {

                        int add_code = MTEAddTable(connection_id, code, code);
                        if (add_code < 0)
                            last_error = ((MicexReturnCode)add_code).ToString();

                    }
                }

                IntPtr[] raw_data = new IntPtr[1];
                int refresh_code = MTERefresh(connection_id, raw_data);
                if (refresh_code < 0)
                    last_error = ((MicexReturnCode)refresh_code).ToString();

                ByteArrayReader reader = ProcessMTEResponse(raw_data);
                if (reader == null)
                {
                    //Log.Push(LogItemType.Error, "Пустой ByteArrayReader при обновлении котировок, connection_id=" + connection_id);
                    return;
                }
                int tables_count = reader.ReadInt();

                List<Quote> pack = null;
                for (int i = 0; i < tables_count; i++)
                {
                    int Ref = reader.ReadInt();
                    ProcessQuotesPack(reader, ref pack);
                }
                if (pack != null)
                    quotes.Enqueue(pack);
            }
            catch (Exception e)
            {
                Log.Push(e, "Ошибка при получении обновления котировок");
            }
            finally
            {
                deamon.ReleaseConnectionId(connection_id);
                is_refreshing_quotes = false;
            }
        }

        public void RefreshOtherTables()
        {
            // Обновляем все остальные таблички
            if (is_refreshing_other_tables)
                return;

            is_refreshing_other_tables = true;
            int connection_id = -1;

            try
            {
                connection_id = deamon.GetConnectionId(MicexConnectionType.RefreshTables);
                if (connection_id < 0)
                    throw new MicexApiException("Не получается получить идентификатор соединения");

                foreach (int code in securities_tables)
                    MTEAddTable(connection_id, code, code);
                foreach (int code in orders_tables)
                    MTEAddTable(connection_id, code, code);
                MTEAddTable(connection_id, trades_table, trades_table);

                if (securities_tables.Count > 0 || orders_tables.Count > 0 || trades_table != 0)
                {
                    IntPtr[] raw_data = new IntPtr[1];
                    int refresh_code = MTERefresh(connection_id, raw_data);
                    if (refresh_code < 0)
                        last_error = ((MicexReturnCode)refresh_code).ToString();

                    ByteArrayReader reader = ProcessMTEResponse(raw_data);
                    if (reader == null)
                    {
                        Log.Push(LogItemType.Error, "Пустой ByteArrayReader при обновлении таблиц, connection_id=" + connection_id); 
                        return;
                    }

                    int tables_count = reader.ReadInt();

                    for (int i = 0; i < tables_count; i++)
                    {
                        int Ref = reader.ReadInt();
                        if (securities_tables.Contains(Ref))
                            ProcessSecurities(reader);
                        else if (orders_tables.Contains(Ref))
                            ProcessOrdersPack(reader);
                        else if (trades_table == Ref)
                            ProcessTradesPack(reader);
                        else
                            throw new ApplicationException("Шлюз выдал ответ с кривым Ref");
                    }
                }
            }
            catch (Exception e)
            {
                Log.Push(e, "Ошибка при получении обновления таблиц от шлюза");
            }
            finally
            {
                deamon.ReleaseConnectionId(connection_id);
                is_refreshing_other_tables = false;
            }
        }

        void ProcessSecurities(ByteArrayReader reader)
        {
            Dictionary<string, Field> structure = fields_converter.GetFieldset(FieldsetName.SecuritiesResponse);

            int strings_quantity = reader.ReadInt();
            for (int j = 0; j < strings_quantity; j++)
            {
                object[] data = fields_converter.Decode(reader, FieldsetName.SecuritiesResponse);
                InstrumentInfo info;

                /// Заполняем объекты InstrumentInfo
                if (sub_type == MicexSubType.Spot)
                {
                    TradingStatus s;
                    switch (data[structure["TRADINGSTATUS"].Index] as string)
                    {
                        case "N": s = TradingStatus.NotAviable; break;
                        case "O": s = TradingStatus.Opening; break;
                        case "C": s = TradingStatus.Closed; break;
                        case "F": s = TradingStatus.Closing; break;
                        case "T": s = TradingStatus.Trading; break;
                        default: s = TradingStatus.NotAviable; break;
                    }

                    if (data[structure["LOTSIZE"].Index] != null && data[structure["SHORTNAME"].Index] != null && data[structure["ISIN"].Index] != null)
                        info = new StockInfo
                        {
                            Name = data[structure["SHORTNAME"].Index] as string,
                            Status = s,
                            LotSize = (int)data[structure["LOTSIZE"].Index],
                            ISIN = data[structure["ISIN"].Index] as string,
                        };
                    else
                        info = null;
                }
                else if (sub_type == MicexSubType.Futures)
                {
                    TradingStatus s;
                    switch (data[structure["TRADINGSTATUS"].Index] as string)
                    {
                        case "N": s = TradingStatus.NotAviable; break;
                        case "O": s = TradingStatus.Opening; break;
                        case "C": s = TradingStatus.Closed; break;
                        case "T": s = TradingStatus.Trading; break;
                        default: s = TradingStatus.NotAviable; break;
                    }

                    string put_call_string = data[structure["PUTCALL"].Index] as string;
                    if (put_call_string == "F")
                        info = new FuturesInfo();
                    else if (put_call_string == "C" || put_call_string == "P")
                    {
                        info = new OptionInfo();
                        (info as OptionInfo).Type = put_call_string == "C" ? OptionType.Call : OptionType.Put;
                    }
                    else
                        info = null;

                    if (info != null)
                    {
                        info.Name = data[structure["SHORTNAME"].Index] as string;
                        info.Status = s;
                        (info as DerivativeInfo).LastTradeDate = (DateTime)data[structure["LASTTRADEDATE"].Index];
                        (info as DerivativeInfo).ExpirationDate = (DateTime)data[structure["LASTDELDATE"].Index];
                        (info as DerivativeInfo).UnderlyingTicker = data[structure["UNDERLYINGSECID"].Index] as string;
                    }
                }
                else
                    throw new NotImplementedException();

                securities.Enqueue(new MicesSecuritiesUpdate
                {
                    SecCode = data[structure["SECCODE"].Index] as string,
                    SecBoard = data[structure["SECBOARD"].Index] as string,
                    Info = info
                });
            }
        }

        void ProcessOrdersPack(ByteArrayReader reader)
        {
            Dictionary<string, Field> structure = fields_converter.GetFieldset(FieldsetName.OrdersResponse);

            int strings_quantity = reader.ReadInt();
            for (int j = 0; j < strings_quantity; j++)
            {
                object[] data = fields_converter.Decode(reader, FieldsetName.OrdersResponse);
                if (data == null)
                    continue;

                // Определяем, есть ли OrderId
                int extref;
                if (!int.TryParse(data[structure["EXTREF"].Index] as string, out extref))
                    extref = 0;

                if (sub_type == MicexSubType.Spot)
                {
                    orders.Enqueue(new MicexOrderUpdate
                    {
                        SecBoard = data[structure["SECBOARD"].Index] as string,
                        SecCode = data[structure["SECCODE"].Index] as string,
                        Account = data[structure["ACCOUNT"].Index] as string,
                        ClientCode = data[structure["CLIENTCODE"].Index] as string,
                        OrderId = (long)data[structure["ORDERNO"].Index],
                        Balance = (int)data[structure["BALANCE"].Index],
                        BuySell = (data[structure["BUYSELL"].Index] as string) == "B" ? Side.Buy : Side.Sell,
                        Price = (double)data[structure["PRICE"].Index],
                        Quantity = (int)data[structure["QUANTITY"].Index],
                        Status = data[structure["STATUS"].Index] as string,
                        OrderTime = (DateTime)data[structure["ORDERTIME"].Index]
                    });
                }
                else if (sub_type == MicexSubType.Futures)
                {
                    orders.Enqueue(new MicexOrderUpdate
                    {
                        SecBoard = data[structure["SECBOARD"].Index] as string,
                        SecCode = data[structure["SECCODE"].Index] as string,
                        Account = String.Empty,
                        ClientCode = data[structure["TRDACCID"].Index] as string,
                        OrderId = (long)data[structure["ORDERNO"].Index],
                        Balance = (int)data[structure["BALANCE"].Index],
                        BuySell = (data[structure["BUYSELL"].Index] as string) == "B" ? Side.Buy : Side.Sell,
                        Price = (double)data[structure["PRICE"].Index],
                        Quantity = (int)data[structure["QUANTITY"].Index],
                        Status = data[structure["STATUS"].Index] as string,
                        OrderTime = (DateTime)data[structure["ORDERTIME"].Index]
                    });
                }
                else
                    throw new NotImplementedException();
            }
        }

        void ProcessTradesPack(ByteArrayReader reader)
        {
            Dictionary<string, Field> structure = fields_converter.GetFieldset(FieldsetName.TradesResponse);

            int strings_quantity = reader.ReadInt();
            for (int j = 0; j < strings_quantity; j++)
            {
                object[] data = fields_converter.Decode(reader, FieldsetName.TradesResponse);
                if (data == null)
                    continue;

                if (sub_type == MicexSubType.Spot)
                {
                    trades.Enqueue(new MicexTradeUpdate
                    {
                        TradeId = (long)data[structure["TRADENO"].Index],
                        ClientCode = data[structure["CLIENTCODE"].Index] as string,
                        SecCode = data[structure["SECCODE"].Index] as string,
                        SecBoard = data[structure["SECBOARD"].Index] as string,
                        Account = data[structure["ACCOUNT"].Index] as string,
                        OrderId = (long)data[structure["ORDERNO"].Index],
                        Price = (double)data[structure["PRICE"].Index],
                        Quantity = (int)data[structure["QUANTITY"].Index],
                        BuySell = (data[structure["BUYSELL"].Index] as string) == "B" ? Side.Buy : Side.Sell,
                        TradeTime = (DateTime)data[structure["TRADETIME"].Index],
                        Commission = (decimal)(double)data[structure["COMMISSION"].Index],
                        Comment = (string)data[structure["BROKERREF"].Index]
                    });
                }
                else if (sub_type == MicexSubType.Futures)
                {
                    trades.Enqueue(new MicexTradeUpdate
                    {
                        TradeId = (long)data[structure["TRADENO"].Index],
                        ClientCode = data[structure["TRDACCID"].Index] as string,
                        SecCode = data[structure["SECCODE"].Index] as string,
                        SecBoard = data[structure["SECBOARD"].Index] as string,
                        Account = String.Empty,
                        OrderId = (long)data[structure["ORDERNO"].Index],
                        Price = (double)data[structure["PRICE"].Index],
                        Quantity = (int)data[structure["QUANTITY"].Index],
                        BuySell = (data[structure["BUYSELL"].Index] as string) == "B" ? Side.Buy : Side.Sell,
                        TradeTime = (DateTime)data[structure["TRADETIME"].Index],
                        Commission = 0,
                        Comment = (string)data[structure["BROKERREF"].Index]
                    });
                }
                else
                    throw new NotImplementedException();
            }
        }
        
        /// <summary>
        /// Разбирает таблицу котировок из ByteArrayReader и помещает обработанные данные в коллекцию
        /// </summary>
        void ProcessQuotesPack(ByteArrayReader reader, ref List<Quote> pack)
        {
            Dictionary<string, Field> structure = fields_converter.GetFieldset(FieldsetName.OrderBookResponse);
            if (pack == null)
                pack = new List<Quote>();

            int strings_quantity = reader.ReadInt();
            for (int j = 0; j < strings_quantity; j++)
            {
                object[] data = fields_converter.Decode(reader, FieldsetName.OrderBookResponse);
                try
                {
                    if (data[structure["BUYSELL"].Index] == null)
                        continue;

                    pack.Add(new Quote
                    {
                        ClassCode = data[structure["SECBOARD"].Index] as string,
                        Ticker = data[structure["SECCODE"].Index] as string,
                        Operation = (data[structure["BUYSELL"].Index] as string) == "B" ? Side.Buy : Side.Sell,
                        Price = (double)data[structure["PRICE"].Index],
                        Volume = (int)data[structure["QUANTITY"].Index],
                        Date = DateTime.Now
                    });
                }
                catch (Exception e)
                {
                    Log.Push(e, "Пришли котировки с некорректными значениями");
                }
            }
        }

        public ServerInfo GetServerInfo()
        {
            int connection_id = -1;
            try
            {
                connection_id = deamon.GetConnectionId(MicexConnectionType.All);
            
                IntPtr[] serv_info = new IntPtr[1];
                var len = new int[1] { -1 };
                var message = new byte[255];
                int code = MTEGetServInfo(connection_id, serv_info, len);

                if (code <= 0)
                {
                    Log.Push(LogItemType.Error, "Ошибка при вызове MTEGetServInfo(): " + ((MicexReturnCode)code));
                    return null;
                }

                var length = len[0];
                var serv_info_data = new byte[length];

                Marshal.Copy(serv_info[0], serv_info_data, 0, length);
                return new ServerInfo(serv_info_data);
            }
            catch (Exception e)
            {
                Log.Push(LogItemType.Error, "Ошибка при вызове GetServerInfo()");
                Log.Push(LogItemType.Error, e.Message);
                return null;
            }
            finally
            {
                deamon.ReleaseConnectionId(connection_id);
            }
        }

        /// <summary>
        /// Читает из заголовка ByteArrayReader всякую ненужную информацию
        /// </summary>
        private ByteArrayReader ProcessMTEResponse(IntPtr[] pmtemsg)
        {
            if (pmtemsg[0] == IntPtr.Zero)
                //throw new CriticalException(pmtemsg.ToString());
                return null;

            var mteMsgHeader = new byte[4];
            Marshal.Copy(pmtemsg[0], mteMsgHeader, 0, 4);
            var bar = new ByteArrayReader(mteMsgHeader);
            int size = bar.ReadInt() + 4;
            var mtemsg = new byte[size];
            Marshal.Copy(pmtemsg[0], mtemsg, 0, size);
            var reader = new ByteArrayReader(mtemsg);
            reader.ReadInt(); // skip size
            return reader;
        }

        private static string ConvertAndDecode(byte[] bytes)
        {
            return Encoding.GetEncoding(1251).GetString(bytes).TrimEnd('\0');
        }

        #region Сигнатуры функций из mtesrl.dll
        [DllImport("mtesrl.dll", CharSet = CharSet.Ansi)]
        private static extern string MTEErrorMsg(int ErrCode);
        
        [DllImport("mtesrl.dll", CharSet = CharSet.Ansi)]
        private static extern int MTEExecTrans(int connectionId, string transactionName, string pars, [MarshalAs(UnmanagedType.LPArray)] byte[] res);

        [DllImport("mtesrl.dll", CharSet = CharSet.Ansi)]
        private static extern int MTEAddTable(int connectionId, int htable, int _ref);

        [DllImport("mtesrl.dll", CharSet = CharSet.Ansi)]
        private static extern int MTECloseTable(int connectionId, int htable);

        [DllImport("mtesrl.dll", CharSet = CharSet.Ansi)]
        private static extern int MTEOpenTable(int connectionId, String tableName, String pars, bool complete, [MarshalAs(UnmanagedType.LPArray)] IntPtr[] pmtemsg);

        [DllImport("mtesrl.dll", CharSet = CharSet.Ansi)]
        private static extern int MTERefresh(int connectionId, [MarshalAs(UnmanagedType.LPArray)] IntPtr[] pmtemsg);

        [DllImport("mtesrl.dll", CharSet = CharSet.Ansi)]
        private static extern int MTEStructure(int connectionId, [MarshalAs(UnmanagedType.LPArray)] IntPtr[] pmtemsg);

        [DllImport("mtesrl.dll", CharSet = CharSet.Ansi)]
        private static extern int MTEGetServInfo(int connectionId, [MarshalAs(UnmanagedType.LPArray)] IntPtr[] servInfo, [MarshalAs(UnmanagedType.LPArray)] int[] len);

        [DllImport("mtesrl.dll", CharSet = CharSet.Ansi)]
        private static extern int MTEGetSnapshot(int connectionId, [MarshalAs(UnmanagedType.LPArray)] IntPtr[] snapshot, [MarshalAs(UnmanagedType.LPArray)] int[] len);

        [DllImport("mtesrl.dll", CharSet = CharSet.Ansi)]
        private static extern int MTESetSnapshot(int connectionId, [MarshalAs(UnmanagedType.LPArray)] byte[] snapshot, int len, [MarshalAs(UnmanagedType.LPArray)] byte[] err);
        #endregion



        public struct MicexOrderUpdate
        {
            public string SecCode;
            public string SecBoard;
            public string Account;
            public string ClientCode;
            public long OrderId;
            public string Status;
            public double Price;
            public int Quantity; /// Объем заявки
            public int Balance;  /// Неисполненный остаток заявки
            public Side BuySell;
            public DateTime? OrderTime; /// Время регистрации заявки в шлюзе
        }

        public struct MicexTradeUpdate
        {
            public long TradeId;
            public string ClientCode;
            public string SecCode;
            public string SecBoard;
            public string Account;
            public long OrderId;
            public double Price;
            public int Quantity;
            public Side BuySell;
            public DateTime TradeTime;
            public string Comment;
            public decimal Commission;
        }

        public struct MicesSecuritiesUpdate
        {
            public string SecCode;
            public string SecBoard;
            public InstrumentInfo Info;
        }
    }

    /// <summary>
    /// Ошибка, возникшая в API ММВБ
    /// </summary>
    public class MicexApiException : ApplicationException
    {
        public MicexApiException()
            : base()
        {
        }

        public MicexApiException(string message)
            : base(message)
        {
        }

        public MicexApiException(string message, Exception inner)
            : base(message, inner)
        {
        }
    }

    enum MicexReturnCode
    {
        Ok = 0,
        Config = -1,
        ServerUnaviable = -2,
        LogError = -3,
        InvalidConnect = -4,
        NotConnected = -5,
        Write = -6,
        Read = -7,
        Tsmr = -8,
        NoMemory = -9,
        Zlib = -10,
        InProgress = -11,
        NotStarted = -12,
        FatalError = -13,
        InvalidHandle = -14,
        DsrOff = -15,
        Unknown = -16,
        BadPtr = -17,
        TransRejected = -18,
        TradeEngineUnaviable = -19,
        NotLoggedIn = -20,
        WrongVersion = -21,
        Logon = -30,
        TooSlowConnect = -31,
        CryptoError = -32,
        TreadError = -33,
        NotImplemented = -34
    }

    public class ServerInfo
    {
        public readonly int ConnectedToMicex;
        public readonly int SessionId;
        public readonly string MicexServerName;
        public readonly byte VersionMajor;
        public readonly byte VersionMinor;
        public readonly byte VersionBuild;
        public readonly byte BetaVersion;
        public readonly byte DebugFlag;
        public readonly byte TestFlag;
        public readonly TimeSpan StartTime;
        public readonly TimeSpan StopTimeMin;
        public readonly TimeSpan StopTimeMax;
        public readonly int NextEvent;
        public readonly DateTime EventDate;
        public readonly string BoardsSelected;
        public readonly string UserId;
        public readonly char SystemId;
        public readonly string ServerIp;

        internal ServerInfo(byte[] servInfoData)
        {
            var reader = new ByteArrayReader(servInfoData);
            ConnectedToMicex = reader.ReadInt();
            SessionId = reader.ReadInt();
            MicexServerName = reader.ReadStringZ(33);
            VersionMajor = reader.ReadByte();
            VersionMinor = reader.ReadByte();
            VersionBuild = reader.ReadByte();
            BetaVersion = reader.ReadByte();
            DebugFlag = reader.ReadByte();
            TestFlag = reader.ReadByte();
            StartTime = reader.ReadTime();
            StopTimeMin = reader.ReadTime();
            StopTimeMax = reader.ReadTime();
            NextEvent = reader.ReadInt();
            EventDate = reader.ReadDate();
            BoardsSelected = reader.ReadStringZ(0);
            UserId = reader.ReadStringZ(13);
            SystemId = reader.ReadChar();
            ServerIp = reader.ReadStringZ(0);
        }
    }
}
